<?php

namespace SGalinski\SgNews\Command;

/***************************************************************
 *  Copyright notice
 *
 *  (c) sgalinski Internet Services (https://www.sgalinski.de)
 *
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 ***************************************************************/

use SGalinski\SgNews\Domain\Model\News;
use SGalinski\SgNews\Domain\Repository\FileReferenceRepository;
use SGalinski\SgNews\Domain\Repository\NewsRepository;
use TYPO3\CMS\Backend\FrontendBackendUserAuthentication;
use TYPO3\CMS\Core\Database\ConnectionPool;
use TYPO3\CMS\Core\Database\Query\Restriction\StartTimeRestriction;
use TYPO3\CMS\Core\DataHandling\DataHandler;
use TYPO3\CMS\Core\Resource\File;
use TYPO3\CMS\Core\Resource\FileInterface;
use TYPO3\CMS\Core\Utility\GeneralUtility;
use TYPO3\CMS\Core\Utility\VersionNumberUtility;
use TYPO3\CMS\Extbase\Mvc\Controller\CommandController;
use TYPO3\CMS\Extbase\Persistence\Exception\IllegalObjectTypeException;
use TYPO3\CMS\Extbase\Persistence\Exception\UnknownObjectException;
use TYPO3\CMS\Extbase\Persistence\Generic\PersistenceManager;
use TYPO3\CMS\Frontend\Controller\TypoScriptFrontendController;

/**
 * Command controller, that migrates data from tx_news to sg_news
 */
class MigrateNewsCommandController extends CommandController {
	/**
	 * @var bool
	 */
	protected $requestAdminPermissions = TRUE;

	/**
	 * @var PersistenceManager
	 */
	protected $persistenceManager;

	/**
	 * @param PersistenceManager $persistenceManager
	 */
	public function injectPersistenceManager(PersistenceManager $persistenceManager) {
		$this->persistenceManager = $persistenceManager;
	}

	/**
	 * @var NewsRepository
	 */
	protected $newsRepository;

	/**
	 * @param NewsRepository $newsRepository
	 */
	public function injectNewsRepository(NewsRepository $newsRepository) {
		$this->newsRepository = $newsRepository;
	}

	/**
	 * @var FileReferenceRepository
	 */
	private $fileReferenceRepository;

	/**
	 * @param FileReferenceRepository $fileReferenceRepository
	 */
	public function injectFileReferenceRepository(FileReferenceRepository $fileReferenceRepository) {
		$this->fileReferenceRepository = $fileReferenceRepository;
	}

	/**
	 * this array maps new pages to their original entry in the tx_news table
	 *
	 * @var array $newsPagesMap
	 */
	protected $newsPagesMap = [];

	/**
	 * this array contains all pages that should be migrated to sg_news
	 *
	 * @var array $pagesToMigrate
	 */
	protected $pagesToMigrate = [];

	/**
	 * @var array $languageMap
	 */
	protected $languageMap = [];

	/**
	 * @var array $languageMap
	 */
	protected $categoryMap = [];

	/**
	 * @param string $copyPageId the page id of the template that should be copied
	 * @param int $categoryPid the page id of the category page
	 * @param int $year only news from that year will be migrated
	 * @param string $languageMapAsJson a json, mapping language ids (old => new). this is needed if the sys_language_uids have changed
	 * @param string $categoryMapAsJson a json, mapping sys_category ids (old => new). t
	 * @param int $pId only news from that pid will be migrated
	 *
	 * @throws IllegalObjectTypeException
	 * @throws UnknownObjectException
	 * @throws \Exception
	 */
	public function runMigrateNewsCommand(
		$copyPageId, $categoryPid, $year = 2015,
		$languageMapAsJson = '{"3":1,"1":0,"2":2,"0":3}',
		$categoryMapAsJson = '{"2":17,"3":16,"4":15,"5":14,"6":14,"7":15,"8":16,"9":17}',
		$pId = 52
	) {
		// fix repair translation bug where tsfe is missing from command controller, can be removed when v1.5 is released
		if (!$GLOBALS['TSFE']) {
			$GLOBALS['TSFE'] = $this->objectManager->get(
				TypoScriptFrontendController::class, $GLOBALS['TYPO3_CONF_VARS'], 0, 0
			);
		}

		$this->languageMap = json_decode($languageMapAsJson, TRUE);
		$this->categoryMap = json_decode($categoryMapAsJson, TRUE);

		$queryBuilder = GeneralUtility::makeInstance(ConnectionPool::class)->getQueryBuilderForTable('tx_news_domain_model_news');
		$queryBuilder->getRestrictions()->removeByType(StartTimeRestriction::class);
		$rows = $queryBuilder->select('*')
			->from('tx_news_domain_model_news')
			->where(
				$queryBuilder->expr()->eq('pid', $queryBuilder->createNamedParameter($pId, \PDO::PARAM_INT))
			)
			->execute()->fetchAll();
		$localDataHandler = GeneralUtility::makeInstance(DataHandler::class);
		$beUser = $this->simulateBackendUser();
		$localCommandMap = [
			'pages' => [
				$copyPageId => [
					'copy' => (int) $categoryPid
				]
			]
		];

		foreach ($rows as $row) {
			// ignore the entry if its not within the given year
			if ((int) date('Y', $row['datetime']) !== $year) {
				continue;
			}

			// if no l10n_parent exists, create a copy of the page
			if ((int) $row['l10n_parent'] === 0) {
				$localDataHandler->start([], $localCommandMap, $beUser);
				$localDataHandler->bypassAccessCheckForRecords = TRUE;
				$localDataHandler->checkModifyAccessList('pages');
				$localDataHandler->process_cmdmap();

				// get the id of the new object
				$newPageId = $localDataHandler->copyMappingArray['pages'][$copyPageId];

				// make entry in news map, so we can fetch it later
				$this->newsPagesMap[$row['uid']] = $newPageId;

				/** @var News $newsPage */
				$newsPage = $this->newsRepository->findByUidIgnoreEnableFields($newPageId);

				if ($newsPage !== NULL) {
					$title = date('Y-m-d', $row['datetime']) . ' - ' . $row['title'];
					$newsPage->setTitle($title);
					$newsPage->setSubtitle($row['title']);

					$date = new \DateTime('@' . $row['datetime']);
					$newsPage->setLastUpdated($date);
					$this->setMatchingTag($row);

					/** @var File $image */
					$file = $this->getMatchingFile($row);
					if ($file instanceof File) {
						$teaserImage1 = $this->fileReferenceRepository->addFileReferenceFromFile(
							$file, $this->newsPagesMap[$row['uid']],
							$this->newsPagesMap[$row['uid']], 'pages', 'tx_sgnews_teaser1_image'
						);

						if ($teaserImage1) {
							$newsPage->addTeaser1Image($teaserImage1);

							$teaserImage2 = $this->fileReferenceRepository->addFileReferenceFromFile(
								$file, $this->newsPagesMap[$row['uid']],
								$this->newsPagesMap[$row['uid']], 'pages', 'tx_sgnews_teaser2_image'
							);
							$newsPage->addTeaser2Image($teaserImage2);
						}
					}

					$this->newsRepository->update($newsPage);
					$this->persistenceManager->persistAll();

					// update content element from the new page
					$queryBuilder = GeneralUtility::makeInstance(ConnectionPool::class)->getQueryBuilderForTable('tt_content');
					$queryBuilder->getRestrictions()->removeAll();
					$queryBuilder->update('tt_content')
						->where(
							$queryBuilder->expr()->andX(
								$queryBuilder->expr()->eq('pid', $queryBuilder->createNamedParameter($newPageId, \PDO::PARAM_INT)),
								$queryBuilder->expr()->eq('sys_language_uid', $queryBuilder->createNamedParameter($this->languageMap[(int) $row['sys_language_uid']], \PDO::PARAM_INT))
							)
						)
						->set('bodytext', $row['bodytext'], TRUE)
						->execute();
				}
			} else { // this row is a translation and should simply update the content accordingly
				$this->updateTranslation($row);
			}
		}
	}

	/**
	 * Get the image file matching the news
	 *
	 * @param array $row
	 * @return null|FileInterface
	 * @throws \Exception
	 */
	private function getMatchingFile(array $row) {
		// match old page id with old file reference id
		$queryBuilder = GeneralUtility::makeInstance(ConnectionPool::class)->getQueryBuilderForTable('sys_file_reference_news_migration');
		$queryBuilder->getRestrictions()->removeAll();
		$fileReferenceResult = $queryBuilder->select('uid', 'uid_local')
			->from('sys_file_reference_news_migration')
			->where(
				$queryBuilder->expr()->andX(
					$queryBuilder->expr()->eq('tablenames', $queryBuilder->createNamedParameter('tx_news_domain_model_news')),
					$queryBuilder->expr()->eq('fieldname', $queryBuilder->createNamedParameter('fal_media')),
					$queryBuilder->expr()->eq('uid_foreign', $queryBuilder->createNamedParameter($row['uid'], \PDO::PARAM_INT))
				)
			)
			->execute()->fetch();
		if (!$fileReferenceResult) {
			return NULL;
		}

		$queryBuilder = GeneralUtility::makeInstance(ConnectionPool::class)->getQueryBuilderForTable('sys_file_news_migration');
		$queryBuilder->getRestrictions()->removeAll();
		$fileResult = $queryBuilder->select('identifier')
			->from('sys_file_news_migration')
			->where(
				$queryBuilder->expr()->eq('uid', $queryBuilder->createNamedParameter($fileReferenceResult['uid_local'], \PDO::PARAM_INT))
			)
			->execute()->fetch();
		if (!$fileResult) {
			return NULL;
		}

		$oldIdentifier = $fileResult['identifier'];
		$resourceFactory = \TYPO3\CMS\Core\Resource\ResourceFactory::getInstance();
		$storage = $resourceFactory->getStorageObject(1);
		if (!$storage->hasFile($oldIdentifier)) {
			return NULL;
		}

		return $storage->getFile($oldIdentifier);
	}

	/**
	 * Get the tag / category, matching the news
	 *
	 * @param array $row
	 */
	private function setMatchingTag(array $row) {
		$queryBuilder = GeneralUtility::makeInstance(ConnectionPool::class)->getQueryBuilderForTable('sys_category_record_mm_news_migration');
		$queryBuilder->getRestrictions()->removeAll();
		$mmRows = $queryBuilder->select('uid_local', 'sorting_foreign')
			->from('sys_category_record_mm_news_migration')
			->where(
				$queryBuilder->expr()->eq('uid_foreign', $queryBuilder->createNamedParameter($row['uid'], \PDO::PARAM_INT))
			)
			->execute()->fetchAll();
		foreach ($mmRows as $mmRow) {
			$values = [
				'uid_local' => $this->categoryMap[(int) $mmRow['uid_local']],
				'uid_foreign' => $this->newsPagesMap[(int) $row['uid']],
				'tablenames' => 'pages',
				'fieldname' => 'tx_sgnews_tags',
				'sorting_foreign' => (int) $mmRow['sorting_foreign']
			];
			$queryBuilder->insert('sys_category_record_mm')
				->values($values);
		}
	}

	/**
	 * Updates content element of a news translationuid
	 *
	 * @param array $row
	 */
	private function updateTranslation(array $row) {
		// get entry in news map
		$parentId = $this->newsPagesMap[$row['l10n_parent']];

		$queryBuilder = GeneralUtility::makeInstance(ConnectionPool::class)->getQueryBuilderForTable('tt_content');
		$queryBuilder->getRestrictions()->removeAll();
		$originalContentElement = $queryBuilder->select('l18n_parent')
			->from('tt_content')
			->where(
				$queryBuilder->expr()->eq('pid', $queryBuilder->createNamedParameter($parentId, \PDO::PARAM_INT))
			)
			->execute()->fetch();

		$queryBuilder->update('tt_content');
		// if its the new default, there is no l18n_parent
		if ((int) $this->languageMap[(int) $row['sys_language_uid'] === 0]) {
			$queryBuilder->where(
				$queryBuilder->expr()->eq('uid', $queryBuilder->createNamedParameter($originalContentElement[0], \PDO::PARAM_INT))
			);
		} else {
			$queryBuilder->where(
				$queryBuilder->expr()->eq('l18n_parent', $queryBuilder->createNamedParameter($originalContentElement[0], \PDO::PARAM_INT))
			);
		}

		// look up the correct language id, if they have changed
		if (isset($this->languageMap[(int) $row['sys_language_uid']])) {
			$queryBuilder->andWhere(
				$queryBuilder->expr()->eq('sys_language_uid', $queryBuilder->createNamedParameter($this->languageMap[(int) $row['sys_language_uid']], \PDO::PARAM_INT))
			);
		} else {
			$queryBuilder->andWhere(
				$queryBuilder->expr()->eq('sys_language_uid', $queryBuilder->createNamedParameter($row['sys_language_uid'], \PDO::PARAM_INT))
			);
		}
		$queryBuilder->set('bodytext', $row['bodytext'], TRUE)
			->execute();

		// possibly the default language needs to be overwritten and the old default translation needs to be preserved
		if (isset($this->languageMap[(int) $row['sys_language_uid']]) && $this->languageMap[(int) $row['sys_language_uid']] === 0) {

			$queryBuilder = GeneralUtility::makeInstance(ConnectionPool::class)->getQueryBuilderForTable('pages');
			$queryBuilder->getRestrictions()->removeAll();
			$result = $queryBuilder->select('title', 'subtitle')
				->from('pages')
				->where(
					$queryBuilder->expr()->eq('uid', $queryBuilder->createNamedParameter($parentId, \PDO::PARAM_INT))
				)
				->execute()->fetch();
			if ($result) {
				$queryBuilder->where(
					$queryBuilder->expr()->eq('sys_language_uid', $queryBuilder->createNamedParameter($this->languageMap[0], \PDO::PARAM_INT))
				);
				if (version_compare(VersionNumberUtility::getCurrentTypo3Version(), '9.0.0', '<')) {
					$queryBuilder->update('pages_language_overlay')
						->andWhere(
							$queryBuilder->expr()->eq('pid', $queryBuilder->createNamedParameter($parentId, \PDO::PARAM_INT))
						);
				} else {
					$queryBuilder->update('pages')
						->andWhere(
							$queryBuilder->expr()->eq('l10n_parent', $queryBuilder->createNamedParameter($this->languageMap[0], \PDO::PARAM_INT))
						);
				}
				$queryBuilder->set('title', $result['title'], TRUE)
					->set('subtitle', $result['subtitle'], TRUE)
					->set('navtitle', '', TRUE)
					->execute();
			}

			$newTitle = date('Y-m-d', $row['datetime']) . ' - ' . $row['title'];
			$queryBuilder->update('pages')
				->where(
					$queryBuilder->expr()->eq('uid', $queryBuilder->createNamedParameter($parentId, \PDO::PARAM_INT))
				)
				->set('title', $newTitle, TRUE)
				->set('subtitle', $row['title'], TRUE)
				->set('lastUpdated', $row['datetime'], TRUE)
				->set('navtitle', '', TRUE)
				->execute();
		} else {
			if (version_compare(VersionNumberUtility::getCurrentTypo3Version(), '9.0.0', '<')) {
				$queryBuilder->update('pages_language_overlay')
					->where(
						$queryBuilder->expr()->eq('pid', $queryBuilder->createNamedParameter($parentId, \PDO::PARAM_INT))
					);
			} else {
				$queryBuilder->update('pages')
					->where(
						$queryBuilder->expr()->eq('l10n_parent', $queryBuilder->createNamedParameter($parentId, \PDO::PARAM_INT))
					);
			}

			// finally translate the page title if necessary
			if (isset($this->languageMap[(int) $row['sys_language_uid']]) && $this->languageMap[(int) $row['sys_language_uid']] > 0) {
				$queryBuilder->andWhere(
					$queryBuilder->expr()->eq('sys_language_uid', $queryBuilder->createNamedParameter($this->languageMap[(int) $row['sys_language_uid']], \PDO::PARAM_INT))
				);
			} else {
				$queryBuilder->andWhere(
					$queryBuilder->expr()->eq('sys_language_uid', $queryBuilder->createNamedParameter($row['sys_language_uid'], \PDO::PARAM_INT))
				);
			}
			$queryBuilder->set('title', date('Y-m-d', $row['datetime']) . ' - ' . $row['title'], TRUE)
				->execute();
		}
	}

	/**
	 * Simulate Backend User for DataHandler
	 *
	 * @return FrontendBackendUserAuthentication
	 */
	private function simulateBackendUser() {
		/** @var \TYPO3\CMS\Backend\FrontendBackendUserAuthentication $BE_USER */
		$BE_USER = GeneralUtility::makeInstance(FrontendBackendUserAuthentication::class);
		$BE_USER->setBeUserByName('admin');
		if ($BE_USER->user['uid']) {
			$BE_USER->fetchGroupData();
		}
		$BE_USER->uc_default['copyLevels'] = '9999';
		$BE_USER->uc = $BE_USER->uc_default;
		$GLOBALS['PAGES_TYPES'][254]['allowedTables'] = '*';
		return $BE_USER;
	}
}
